/*
 * ----------------------------------------------------------------------------
		**************** SER2IEC V2.0 ***************************
		by Yorck Thiele
		ATMEGA1284 @14,7456MHz
 * ----------------------------------------------------------------------------
 *
 */

#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
	
#define PWR_ON_MSG "SER2IEC V2.0"
#define IOPORT PORTD
#define FAST_SER_PORT_D DDRB
#define FAST_SER_PORT_I PINB
#define IECPORT_O PORTA
#define IECPORT_I PINA
#define IECPORT_D DDRA

#define FS_DATA PB5
#define IEC_SRQ PB7

#define IEC_RESET PA0
#define IEC_ATN PA1
#define IEC_CLK PA2
#define IEC_DATA PA3
#define TRIGGER PA4
#define IEC_ATN1 PA5

#define LED1 PD5
#define LED2 PD6

#define IEC_EOI 6

#define CBM1541 1
#define CBM1571 2
#define CBM1581 3|128

#define NOT_TIMEOUT_T0 !(TIFR0 & (1<<TOV0)) //==0

#define NOT_TIMEOUT_T1 !(TIFR1 & (1<<TOV1)) //==0
#define TIMEOUT_T1 (TIFR1 & (1<<TOV1)) //!=0

#define SPI_BYTE_RDY (SPSR & (1<<SPIF))
#define NOT_SPI_BYTE_RDY !(SPSR & (1<<SPIF))



volatile uint8_t gcr1;
volatile uint8_t gcr2;
volatile uint8_t rxbuffer[260];
//volatile uint8_t tx_rx_buffer[340];
volatile uint8_t sectorbuffer[11000];
//volatile uint16_t tx_rx_buffer_ptr;
//volatile uint8_t rxbuff;
volatile uint16_t bytes2pc;

/*
volatile uint8_t sec_buff_ptr;
volatile uint8_t rxbuffindex;

volatile uint8_t frameerror;
*/
volatile uint8_t pvalid;
volatile uint8_t pid;
volatile uint8_t plen;

volatile uint8_t iec_status;
volatile uint8_t iec_eoi; // EOI flag for IEC output
volatile uint8_t iec_dev_adr; // DEVICE ADDRESS
volatile uint8_t iec_sec_adr; // SECONDARY ADDRESS
volatile uint8_t listentalk; // flag, 0=LISTEN, 1=TALK, 2=UNLISTEN, 3=UNTALK
volatile uint8_t byteinbuff; // flag, onebytebuff is occupied
volatile uint8_t onebytebuff;// buffer for one byte to send
volatile uint8_t fastserial; // FSM (fast serial mode) flag
volatile uint8_t fastdevice; // FSM drive if set, cleared by 'IEC_open'
volatile uint8_t temp;


const uint8_t pgmLOADER[] PROGMEM =
{
	//40 bytes loader for 1541
	0x4D, 0x2D, 0x45, 0x05, 0x02, 0xE8, 0x8E, 0x0B, 0x18, 0x8C, 0x08, 0x18, 0x8C, 0x0A, 0x18, 0x20,
	0xC2, 0xDE, 0xAD, 0x0A, 0x18, 0xD0, 0x0E, 0x8A, 0x2C, 0x0D, 0x18, 0xF0, 0xFB, 0xAD, 0x0A, 0x18,
	0x91, 0x6F, 0xC8, 0xD0, 0xF2, 0x4C, 0x1B, 0xCB
};

const uint8_t pgmLOADER7X[] PROGMEM =
{
	//19 bytes loader for 157x
	0x4D, 0x2D, 0x45, 0x06, 0x02, 0x01, 0xA9, 0x20, 0x85, 0x3B, 0x8C, 0x74, 0x02, 0x20, 0x12, 0x84,
	0x4C, 0x00, 0x03
};
//===================================================================================
ISR(USART1_RX_vect)
{
  
  static uint16_t sec_buff_ptr;
  static uint8_t rxbuffindex;
  static uint8_t frameerror;
  static uint8_t pages;

  uint8_t serbyte;


  if (bit_is_clear(UCSR1A, FE1))
    {
	  serbyte = UDR1;
	  if (frameerror!=0) {rxbuffindex=0;}
	  switch(rxbuffindex)
	  {
	   case 0:
	    pid=serbyte;
		sec_buff_ptr=0;
		plen=0;
		break;
	   case 1:
	    plen=serbyte;
		break;
       case 2:
	    pages=serbyte;
		break;
       case 3:
		  if(plen==0)
	      {
		   sectorbuffer[sec_buff_ptr++]=serbyte;
		   if((sec_buff_ptr/256)==pages) pvalid=1; // pages x 256 bytes received?
		  }
		  break;
	   }
	   if((plen!=0) || (rxbuffindex<3))
		 {
		  rxbuffer[rxbuffindex++]=serbyte;
		  if(rxbuffindex==plen) pvalid=1;
		 }
      frameerror=0;
	}
	else
	{
	 serbyte = UDR1;
     if (frameerror!=0) {wdt_enable(WDTO_15MS);for(;;);} // 2 x FRAME ERROR --> RESET ATMEGA
	 frameerror=1;
	}
}
//===================================================================================
void
delay_n(uint16_t n)
{
//n x 0.54us
//32ms max
TCNT1=65535-n;
TIFR1=(1<<TOV1);
while (NOT_TIMEOUT_T1);
}
//===================================================================================
void
delay_us(uint16_t usec)
{
//n x 1.085us (n>1)!!
//32ms max
TCNT1=65535-usec*2+2;
TIFR1=(1<<TOV1);
while (NOT_TIMEOUT_T1);
}
//===================================================================================
void
delay_ms(uint16_t msec)
{
while (msec != 0)
{
 delay_us(1000);
 msec--;
}
}
//===================================================================================
void
set_timeout(uint16_t usec)
{
//n x 1.085us
//32ms max
TCNT1=65535-usec*2;
TIFR1=(1<<TOV1);
}
//===================================================================================
void
send_byte_to_PC(uint8_t byte)
{
  while (!(UCSR1A & (1<<UDRE1)));
  UDR1=byte;
  bytes2pc++;
}
//===================================================================================
void
frame_termination(uint16_t returnbytes)
{
uint16_t payload=bytes2pc;

while(bytes2pc < returnbytes)
 {
  send_byte_to_PC(0);
 }
send_byte_to_PC(payload & 255);
send_byte_to_PC(payload / 256);
send_byte_to_PC(iec_status);
send_byte_to_PC(pid);
}
//===================================================================================
void
clear_spif(void)
{
temp=SPSR;// clear receive complete flag (SPIF)
temp=SPDR;
}
//===================================================================================
void
fastbus_input(void)
{
SPCR=(1<<SPE) | (1<<CPOL) | (1<<CPHA); // set SPI as slave input

clear_spif();
}
//===================================================================================
void
fastbus_output(void)
{
SPCR=0; // disable SPI
clear_spif();
}
//===================================================================================
void
send_byte_fast1(uint8_t fastbyte)
{
uint8_t i;
//CLK high period min. 1us for 6526@2MHz!!
//CLK: LOW=750ns;HIGH=1250ns
//16us per byte

for (i=0x80;i!=0;i>>=1) //send 8 bits, MSB first
{
 FAST_SER_PORT_D |= (1<<IEC_SRQ); // SRQ=0 (FAST_CLK)
 if ((fastbyte & i) != 0) IECPORT_D &= ~(1<<IEC_DATA); else IECPORT_D |= (1<<IEC_DATA);

 FAST_SER_PORT_D &= ~(1<<IEC_SRQ); // SRQ=1 (FAST_CLK)

 TIFR1=(1<<TOV1);//waste some ns
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 
 TIFR1=(1<<TOV1);



}

IECPORT_D &= ~(1<<IEC_DATA); // DATA=1


}
//===================================================================================
//CLK high period >1us for 6526@2MHz!!
//timing for CLK: LOW=700ns;HIGH=1500ns
void
send_byte_fast(uint8_t fastbyte)
{
	uint8_t i;


for (i=0x80;i!=0;i>>=1) //send 8 bits, MSB first
{
	FAST_SER_PORT_D |= (1<<IEC_SRQ); // SRQ=0 (FAST_CLK)
	if ((fastbyte & i) != 0) IECPORT_D &= ~(1<<IEC_DATA); else IECPORT_D |= (1<<IEC_DATA);
	FAST_SER_PORT_D &= ~(1<<IEC_SRQ); // SRQ=1 (FAST_CLK)
	TCNT1=65534;
	TIFR1=(1<<TOV1);
	while (NOT_TIMEOUT_T1);

}

IECPORT_D &= ~(1<<IEC_DATA); // DATA=1


}
//===================================================================================
void
send_byte_fast_alt(uint8_t fastbyte)
{
uint8_t i;
//CLK high period min. 1us for 6526@2MHz!!
//16us per byte

for (i=0x80;i!=0;i>>=1) //send 8 bits, MSB first
{
 FAST_SER_PORT_D |= (1<<IEC_SRQ); // SRQ=0 (FAST_CLK)
 if ((fastbyte & i) != 0) IECPORT_D &= ~(1<<IEC_DATA); else IECPORT_D |= (1<<IEC_DATA);
 //delay_us(3);
//


 FAST_SER_PORT_D &= ~(1<<IEC_SRQ); // SRQ=1 (FAST_CLK)
 TIFR1=(1<<TOV1);//waste some ns
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);
 TIFR1=(1<<TOV1);

}

IECPORT_D &= ~(1<<IEC_DATA); // DATA=1

}
//===================================================================================
void
send_string_RS232(char *txtptr)
{
uint8_t z;

for (z=0;txtptr[z]!=0;z++)
{
 send_byte_to_PC(txtptr[z]); // send byte via USART
}
}
//===================================================================================
void
init(void)
{
cli();
wdt_reset();
MCUSR &= ~(1<<WDRF);
wdt_disable();
WDTCSR |= (1<<WDCE) | (1<<WDE); 
WDTCSR=0;


DDRD = (1 << LED1) | (1 << LED2);
IECPORT_D = (1 << TRIGGER);
//Timer0
TCCR0B=(1 << CS00);// CLK (68ns) for Timer0

//Timer1
TCCR1B=(1 << CS11);// CLK/8 (0.54us) for Timer1 (overflow @ 35.67ms)


//UART
UBRR1H=0;
UBRR1L=1; //460800 Baud @ 14.7456MHz
UCSR1C = (1<<UCSZ11) | (1<<UCSZ10); // no parity, 1 stop bit, 8 data bits
UCSR1B = (1<<RXEN1) | (1<<TXEN1) | (1<<RXCIE1);

delay_ms(1);

send_string_RS232(PWR_ON_MSG);
sei(); //enable RX interrupt

}
//===================================================================================
int
get_clk(void)
{
uint8_t p;

do{p=IECPORT_I & (1<<IEC_CLK);}

while (p != (IECPORT_I & (1<<IEC_CLK)));
return p;

}
//===================================================================================
int
get_data(void)
{
uint8_t p;

do{p=IECPORT_I & (1<<IEC_DATA);}

while (p != (IECPORT_I & (1<<IEC_DATA)));
return p;

}
//===================================================================================
void
iec_reset()
{
uint16_t timeout=2200;//Timeout 2s

IECPORT_D = (1<<IEC_RESET) |(1<<TRIGGER); //RESET=0; CLK=1; DATA=1
delay_ms(200); // 200ms delay
IECPORT_D &= ~(1<<IEC_RESET); //RESET=1
//wait for response on DATA (1541->1.2s) or Timeout
while ((timeout!=0) && (get_data()==0))
{
 delay_ms(1);
 timeout--;
}

if(timeout==0) iec_status=255;//Timeout 2s

}
//===================================================================================
void
iec_terminate(void)
{
IECPORT_D &= ~(1<<IEC_ATN); //ATN=1
delay_us(60); //TR FRAME TO RELEASE OF ATN
IECPORT_D &= ~(1<<IEC_CLK); // CLK=1
IECPORT_D &= ~(1<<IEC_DATA);// DATA=1
}
//===================================================================================
int
iec_byte_input(void)
{
uint8_t iecb=0,i;

IECPORT_D &= ~(1<<IEC_CLK); // CLK=1
delay_us(5);
while(get_clk()==0); // wait for "TALKER READY-TO-SEND"
set_timeout(256); //EOI-TIMEOUT
// don't use delay_us() during timout seqence !!!
IECPORT_D &= ~(1<<IEC_DATA); //DATA=1, set "LISTENER READY FOR DATA"
while ((NOT_TIMEOUT_T1) && (get_clk()!=0)); // wait for "TALKER SENDING"

if (TIMEOUT_T1) //EOI-TIMEOUT triggered
{
 iec_status = (1<<IEC_EOI); //EOI, last byte
 IECPORT_D |= (1<<IEC_DATA); //DATA=0, set "EOI-TIMEOUT HANDSHAKE"
 delay_us(60); //TEI EOI RESPONSE HOLD TIME
 set_timeout(256);
// don't use delay_us() during timout seqence !!!
 IECPORT_D &= ~(1<<IEC_DATA); //DATA=1, set "LISTENER READY FOR DATA"
 while ((NOT_TIMEOUT_T1) && (get_clk()!=0)); // wait for last byte "TALKER SENDING"
 if (TIMEOUT_T1) {iec_status |= 2; fastserial=0; return 0;}// TIME OUT !!!
}

do
{
 if(SPI_BYTE_RDY) //byte received via fastbus?
 {
 iecb=SPDR;
 fastserial=1;
 fastdevice=1;
 goto byte_ready;
 } 
}
while(get_clk()==0);

for (i=1;i!=0;i<<=1)
{
 delay_us(5);
 while(get_clk()==0); // wait for CLK 0->1 transition "DATA VALID"
 delay_us(5);
 if (get_data() != 0) iecb |= i;
 while(get_clk() != 0);// wait for CLK 1->0 transition
}

byte_ready:
IECPORT_D |= (1<<IEC_DATA); //DATA=0, set "LISTENER DATA ACCEPTED"
if (iec_status == (1<<IEC_EOI)) delay_us(70); // TFR EOI ACKNOWLEDGE

return iecb;

}
//===================================================================================
void
iec_byte_output(uint8_t iecbyte)
{
uint8_t i;

 IECPORT_D &= ~(1<<IEC_DATA); // DATA=1
 delay_us(5);
 if (get_data()!=0)
 {
  iec_status = 0x80; // "DEVICE NOT PRESENT"
  goto iec_out_err;
 }
 IECPORT_D &= ~(1<<IEC_CLK); // CLK=1, set "TALKER READY TO SEND"
 if(iec_eoi !=0)
 {
  while(get_data() == 0); // wait for DATA 0->1 transition "LISTENER READY FOR DATA"
  delay_us(5);
  // trigger EOI-TIMEOUT:
  while(get_data() != 0); // wait for DATA 1->0 transition "EOI-TIMEOUT HANDSHAKE"
 }
 delay_us(5);
 do
 {
   if(SPI_BYTE_RDY)  //DRF (Drive Request Fast) received?
   {
   fastserial=1;
   fastdevice=1;
   clear_spif();
   }
 } while(get_data() == 0); // wait for DATA 0->1 transition "LISTENER READY FOR DATA"

 delay_us(60); //TNE NON-EOI RESPONSE
 IECPORT_D |= (1<<IEC_CLK); // CLK=0, set "TALKER SENDING"

if((fastserial==1) && ((IECPORT_D & (1<<IEC_ATN))==0)) 
{
 fastbus_output();
 send_byte_fast(iecbyte);
 fastbus_input();
}
else
{
 if (get_data()==0)
 {
  iec_status = 0x03; // "TIME OUT"
  goto iec_out_err;
 }
 for (i=1;i!=0;i<<=1) //send 8 bits, LSB first
 {
  delay_us(45);
  if ((iecbyte & i) != 0) IECPORT_D &= ~(1<<IEC_DATA); else IECPORT_D |= (1<<IEC_DATA);
  delay_us(26); //TS: BIT SET-UP TALKER
  IECPORT_D &= ~(1<<IEC_CLK); // CLK=1
  delay_us(26); //TV: DATA VALID
  IECPORT_D &= ~(1<<IEC_DATA);// DATA=1
  IECPORT_D |= (1<<IEC_CLK); // CLK=0
 }
}

set_timeout(1000);//TF: FRAME HANDESHAKE
// don't use delay_us() during timout seqence !!!

while ((NOT_TIMEOUT_T1) && (get_data()!=0)); // wait for"LISTENER DATA ACCEPTED"


if (TIMEOUT_T1)
 {
  iec_status = 0x03; // "TIME OUT"
iec_out_err:
  fastserial=0;
  iec_terminate();
 }

delay_us(100); //TBB BETWEEN BYTES TIME
}
//===================================================================================
void
iec_command(void)
{
uint8_t cmdbyte=0;

if (byteinbuff != 0)
 {
  iec_eoi=1;
  iec_byte_output(onebytebuff);
  byteinbuff=0;
  iec_eoi=0;
 }

switch (listentalk)
  {
   case 0: //LISTEN
	cmdbyte = iec_dev_adr | 0x20;
	break;
   case 1: //TALK
	cmdbyte = iec_dev_adr | 0x40;
	break;
   case 2: //UNLISTEN
    cmdbyte = 0x3f;
	fastserial=0;
	break;
   case 3: //UNTALK
    cmdbyte = 0x5f;
	fastserial=0;
	delay_us(100);
	break;
  }
 

switch (listentalk)
{
	case 0: //LISTEN
	case 1: //TALK
	fastbus_output();
	send_byte_fast(0xFF); // *1 HRF (Host Request Fast)
	fastbus_input();
	delay_us(120); //C128 -->120us
	break;
}


IECPORT_D |= (1<<IEC_ATN);// ATN=0
delay_us(5);
IECPORT_D |= (1<<IEC_CLK);// CLK=0
IECPORT_D &= ~(1<<IEC_DATA);// DATA=1

delay_us(1000);// TAT ATN Response

iec_byte_output(cmdbyte);

}
//===================================================================================
void
iec_secadr(uint8_t sadr)
{
	
	IECPORT_D |= (1<<IEC_ATN);// ATN=0
	delay_us(5);
	IECPORT_D |= (1<<IEC_CLK);// CLK=0
	IECPORT_D &= ~(1<<IEC_DATA);// DATA=1

	delay_us(1000);// TAT ATN Response
	iec_byte_output(sadr);

	switch (listentalk)
	{
		case 0: //LISTEN
		IECPORT_D &= ~(1<<IEC_ATN);// ATN=1
		break;
		
		case 1: //TALK
		if ((iec_status & 0x80)==0) // TALK: TALK-ATTENTION TURNAROUND
		{
			IECPORT_D |= (1<<IEC_DATA);// DATA=0
			IECPORT_D &= ~(1<<IEC_ATN);// ATN=1
			delay_us(50); //TK TALK-ATTENTION RELEASE
			IECPORT_D &= ~(1<<IEC_CLK); // CLK=1
			delay_us(50);//important!!
			while(get_clk() != 0); // wait for CLK=0, TALK-ATTENTION TURNAROUND finished
			//delay_us(400);
		}
		else
		{
			iec_terminate();
		}
		break;
	}

}
//===================================================================================
void
iec_listen(void)
{
listentalk=0;
iec_command();
}
//===================================================================================
void
iec_talk(void)
{
listentalk=1;
iec_command();
}
//===================================================================================
void
iec_unlisten(void)
{
if((IECPORT_I & (1<<IEC_ATN))==0) iec_status=4;// error #4, if ATN=L
else
 {
 listentalk=2;
 iec_command();
 iec_terminate();
 }
}
//===================================================================================
void
iec_untalk(void)
{
if((IECPORT_I & (1<<IEC_ATN))==0) iec_status=4;// error #4, if ATN=L
else
 {
 listentalk=3;
 iec_command();
 iec_terminate();
 }
 
}

//===================================================================================
void
iec_set_talker(void)
{
if((IECPORT_I & (1<<IEC_ATN))==0) iec_status=4;// error #4, if ATN=L
else
 {
 iec_dev_adr=rxbuffer[2]; 
 iec_sec_adr=rxbuffer[3] ;
 iec_talk();
 iec_secadr(iec_sec_adr | 0x60); // device as talker
 }
}
//===================================================================================
void
iec_set_listener(void)
{
if((IECPORT_I & (1<<IEC_ATN))==0) iec_status=4;// error #4, if ATN=L
else
 {
 iec_dev_adr=rxbuffer[2] & 0x1f;
 iec_sec_adr=rxbuffer[3] ;
 iec_listen();
 iec_secadr(iec_sec_adr | 0x60); // device as listener
 }
}
//===================================================================================
void
iec_send_byte(uint8_t iecbyte)
{
if(byteinbuff==0) byteinbuff=1; else iec_byte_output(onebytebuff);
onebytebuff=iecbyte;
}
//===================================================================================
void
iec_send_string(void)
{
uint8_t i;

i=4;
while((i<plen) && (iec_status)==0)
{
iec_send_byte(rxbuffer[i++]);
}
iec_unlisten();
 }
//===================================================================================
void
iec_open(void)
{
iec_status=0;
fastdevice=0;
//IOPORT &= ~(1<<LED1);
if((IECPORT_I & (1<<IEC_ATN))==0) iec_status=4;// error #4, if ATN=L
else
 {
iec_dev_adr=rxbuffer[2];
iec_sec_adr=rxbuffer[3];
iec_listen(); 
iec_secadr(iec_sec_adr | 0xF0);// OPEN
iec_send_string();
}
if(fastdevice) iec_status =0x20;

}
//===================================================================================
void
iec_close(void)
{
iec_dev_adr=rxbuffer[2]; 
iec_sec_adr=rxbuffer[3] ;
iec_listen(); 
iec_secadr(iec_sec_adr | 0xE0);// CLOSE
iec_unlisten();

}
//===================================================================================
void
iec_receive_string(uint16_t readmaxlen)
{

iec_set_talker();

while((readmaxlen!=0) && (iec_status==0))
{
 // send byte via USART
 send_byte_to_PC(iec_byte_input());

 readmaxlen--;
}


iec_untalk();

}
//===================================================================================
void
send_cmd(char *txtptr)
{
uint8_t z;

for (z=0;txtptr[z]!=0;z++)
{
 if(iec_status==0) iec_send_byte(txtptr[z]);
}
}
//===================================================================================
void
memory_read_1581(uint8_t mem_start_hi, uint8_t pages)
{
	uint16_t n;
	uint8_t timeouts=60; //2s
	
	iec_set_listener(); send_cmd("U0>MR");
	iec_send_byte(mem_start_hi); iec_send_byte(pages);
	iec_unlisten();
	if(iec_status==0)
	{
		fastbus_input(); // set SPI as slave input
		for(n=256*pages;n>0;n--)
		{
			clear_spif();
			IECPORT_D ^= (1<<IEC_CLK);
			while(NOT_SPI_BYTE_RDY && timeouts!=0) // wait for status byte
			{
				if(TIMEOUT_T1) {timeouts--;TIFR1=(1<<TOV1);}
			}
			send_byte_to_PC(SPDR);
		}
		IECPORT_D &= ~(1<<IEC_CLK); // CLK=1
		if (timeouts==0) {iec_status = 2; return;}// TIME OUT !!!
	}
}
//===================================================================================
// +++++++++++++++ SPI receiver routine II +++++++++++++++++++
// uses SPI interface (SCK & MOSI pins)
// 
void
iec_receive_track1581(void)
{
//uint16_t n;
uint8_t status, timeouts=30; //track read status

fastbus_input(); // set SPI as slave input
IECPORT_D |= (1<<IEC_CLK); // CLK=0


while(NOT_SPI_BYTE_RDY && timeouts!=0) // wait for status byte
{
 if(TIMEOUT_T1) {timeouts--;TIFR1=(1<<TOV1);}
}
IECPORT_D &= ~(1<<IEC_CLK); // CLK=1
if (timeouts==0) {iec_status = 2; return;}// TIME OUT !!!

status=SPDR;
send_byte_to_PC(SPDR);


if((status & 0x8f)==0) //error?
{
	memory_read_1581(0x0c,20);//read memory $0C00-$1FFF (20 pages)
	/*
 iec_set_listener(); send_cmd("U0>MR"); //read memory $0C00-$1FFF (20 pages)
 iec_send_byte(0x0c); iec_send_byte(20);
 iec_unlisten();
 for(n=256*20;n>0;n--)
  {
  clear_spif();
  IECPORT_D ^= (1<<IEC_CLK);
  while(NOT_SPI_BYTE_RDY && timeouts!=0) // wait for status byte
  {
   if(TIMEOUT_T1) {timeouts--;TIFR1=(1<<TOV1);}
  }
  send_byte_to_PC(SPDR);
  }
IECPORT_D &= ~(1<<IEC_CLK); // CLK=1
*/
}
if (TIMEOUT_T1) {iec_status = 2; return;}// TIME OUT !!!

}
//===================================================================================
// +++++++++++++++ BURST WRITE routine +++++++++++++++++++
//
// uses SPI
// sector packet is 256 bytes 
// 
void
iec_burst_write(uint8_t nostatus)
{
uint16_t n;
uint8_t CLKTOGGLE=0;

//CLKTOGGLE=(IECPORT_I & (1<<IEC_CLK))^(1<<IEC_CLK);

//start transmission
fastbus_output();

for(n=0;n<256;n++)
{
 // wait for CLK TOGGLE
 while ((IECPORT_I & (1<<IEC_CLK))!=CLKTOGGLE);

 send_byte_fast(sectorbuffer[n]);
 CLKTOGGLE^=(1<<IEC_CLK);
}
fastbus_input(); // set SPI as slave input
if(nostatus==0)
{
 IECPORT_D |= (1<<IEC_CLK); // CLK=0
 while(NOT_SPI_BYTE_RDY); // wait for statusbyte
 send_byte_to_PC(SPDR); // send statusbyte via USART
}
IECPORT_D &= ~(1<<IEC_CLK); // CLK=1

}
//===================================================================================
// +++++++++++++++ BURST WRITE routine +++++++++++++++++++
//
// uses SPI
// track packet is 2x5120 bytes (40 sectors)
// 
void
iec_burst_write_track()
{
uint16_t sec_buff_ptr=0;
uint8_t CLKTOGGLE=0,sectors;

//CLKTOGGLE=(IECPORT_I & (1<<IEC_CLK))^(1<<IEC_CLK);

for(sectors=0;sectors<40;sectors++)
{
//start transmission
 fastbus_output();

 do
 {
  // wait for CLK TOGGLE
  while ((IECPORT_I & (1<<IEC_CLK))!=CLKTOGGLE);
  send_byte_fast(sectorbuffer[sec_buff_ptr++]);
  CLKTOGGLE^=(1<<IEC_CLK);
 }
 while((sec_buff_ptr & 255) !=0);

 fastbus_input(); // set SPI as slave input
 
 IECPORT_D |= (1<<IEC_CLK); // CLK=0
 while(NOT_SPI_BYTE_RDY); // wait for statusbyte
 IECPORT_D &= ~(1<<IEC_CLK); // CLK=1

 if((SPDR & 0x8F)>1) break; //controller error?
}

send_byte_to_PC(SPDR); // send statusbyte via USART
}
//===================================================================================
void
send_byte_fast_to_6522(uint8_t fastbyte)
{
uint8_t i, n, bit;

	
fastbus_output();

if(iec_status==0)
{
 for (i=0x80;i!=0;i>>=1) //send 8 bits, MSB first
  {
   if ((fastbyte & i) != 0) bit=IECPORT_D & ~(1<<IEC_DATA); else bit=IECPORT_D | (1<<IEC_DATA);
   while((FAST_SER_PORT_I & (1<<IEC_SRQ))!=0); // wait for SRQ=0 (FAST_CLK)
   IECPORT_D=bit; //set DATA
   n=0;
   while(n<3){if((FAST_SER_PORT_I & (1<<IEC_SRQ))!=0) n++;}; // wait for stable SRQ=1 (FAST_CLK)
  }
 IECPORT_D &= ~(1<<IEC_DATA); // DATA=1
}

}
//===================================================================================
int
receive_fast_byte(uint16_t timeout_ms)
{

fastbus_input();
timeout_ms=timeout_ms/36;
while(NOT_SPI_BYTE_RDY && (timeout_ms>0))
{
 while (NOT_SPI_BYTE_RDY && NOT_TIMEOUT_T1); //Timer1 timeout 35.67ms
 if(TIMEOUT_T1) {timeout_ms--;TIFR1=(1<<TOV1);}
}
if(NOT_SPI_BYTE_RDY) iec_status = 2; // TIME OUT !!! (Fast Bus SRQ problem)
return SPDR;

}

//===================================================================================
void
send_buffer_to_slow_device()
{
uint8_t i=0;

while((i<41) && (iec_status)==0) //transfer LOADER to floppy input buffer
{
 iec_send_byte(pgm_read_byte(&pgmLOADER[i++]));
}
iec_unlisten();

//check LOADER response
IECPORT_D |= (1<<IEC_DATA); // DATA=0
receive_fast_byte(100);
//receive_fast_byte(1);
IECPORT_D &= ~(1<<IEC_DATA); // DATA=1
i=0;
if(iec_status==0)
{
 do 
 {//buffer upload to floppy RAM $300-$3FF and execute $300
    send_byte_fast_to_6522(sectorbuffer[i++]);
 }while(i!=0);
 send_byte_fast_to_6522(255); //timing dummy
}
else 
{
 iec_reset();
}
}
//===================================================================================
void
send_cmd_to_slow_device()
{
uint8_t i;


IECPORT_D |= (1<<IEC_CLK); // CLK=0
send_byte_fast_to_6522(0x55); //handshake dummy
IECPORT_D &= ~(1<<IEC_CLK); // CLK=1

if(iec_status==0)
{
 i=2;
 while((i<8) && (iec_status==0)) //send 6 bytes
 {
  send_byte_fast_to_6522(rxbuffer[i++]);
 }
send_byte_fast_to_6522(255); //timing dummy
}
}
//===================================================================================
void
read_track(uint16_t bytes)
{
uint16_t result=0;
uint8_t timeouts=0, flag=0;



//Timer1 timeout 35.67ms

fastbus_input();

while(bytes!=0 && (timeouts<60))
{
 while (NOT_SPI_BYTE_RDY && NOT_TIMEOUT_T1);
 if((SPSR & (1<<SPIF)))
 {
  if(flag==0){flag++;TCNT1=0;TIFR1=(1<<TOV1);timeouts=0;} //first byte received?
  send_byte_to_PC(SPDR);
  bytes--;
  if(timeouts==0) result++; // count bytes until Timer1 timeout 35.67ms
 }
 if(TIMEOUT_T1) {timeouts++;TIFR1=(1<<TOV1);}
}
send_byte_to_PC(result & 255);
send_byte_to_PC(result / 256);
if(bytes!=0) {iec_status = 8;}//Timeout

}
//===================================================================================
//CLK high period >1us for 6526@2MHz!!
//timing for CLK: LOW=750ns;HIGH=1600ns
//see also: send_byte_fast();
void
write_track_fast_device()
{
uint16_t sec_buff_ptr=0;
uint8_t  byte, i, clk_toggle=0;

fastbus_output();
do
{

 while (get_clk()!=clk_toggle); //wait for clk toggle
 byte=sectorbuffer[sec_buff_ptr++];
 
 
 for (i=0x80;i!=0;i>>=1) //send 8 bits, MSB first
 {
	FAST_SER_PORT_D |= (1<<IEC_SRQ); // SRQ=0 (FAST_CLK)
	if ((byte & i) != 0) IECPORT_D &= ~(1<<IEC_DATA); else IECPORT_D |= (1<<IEC_DATA);
	FAST_SER_PORT_D &= ~(1<<IEC_SRQ); // SRQ=1 (FAST_CLK)
	TCNT1=65534;
	TIFR1=(1<<TOV1);
	while (NOT_TIMEOUT_T1);
 }

 IECPORT_D &= ~(1<<IEC_DATA); // DATA=1

 clk_toggle^=(1<<IEC_CLK);
}while(byte!=0);

fastbus_input(); // set SPI as slave input

}
//===================================================================================
void
write_track_slow_device()
{
uint16_t sec_buff_ptr=0;
uint8_t  byte, bit;

fastbus_output();

if(iec_status==0)
{
 do  //send 9 bits, MSB first
 {
 byte=sectorbuffer[sec_buff_ptr++];
 if ((byte & 0x80) != 0) bit=IECPORT_D & ~(1<<IEC_DATA); else bit=IECPORT_D | (1<<IEC_DATA);
 while((FAST_SER_PORT_I & (1<<IEC_SRQ))!=0); // wait for SRQ=0 (FAST_CLK)
 IECPORT_D=bit; //set DATA bit 7
 if ((byte & 0x40) != 0) bit=IECPORT_D & ~(1<<IEC_DATA); else bit=IECPORT_D | (1<<IEC_DATA);
 while((FAST_SER_PORT_I & (1<<IEC_SRQ))==0); // wait for SRQ=1 (FAST_CLK)
 while((FAST_SER_PORT_I & (1<<IEC_SRQ))!=0); // wait for SRQ=0 (FAST_CLK)
 IECPORT_D=bit; //set DATA bit 6
 if ((byte & 0x20) != 0) bit=IECPORT_D & ~(1<<IEC_DATA); else bit=IECPORT_D | (1<<IEC_DATA);
 while((FAST_SER_PORT_I & (1<<IEC_SRQ))==0); // wait for SRQ=1 (FAST_CLK)
 while((FAST_SER_PORT_I & (1<<IEC_SRQ))!=0); // wait for SRQ=0 (FAST_CLK)
 IECPORT_D=bit; //set DATA bit 5
 if ((byte & 0x10) != 0) bit=IECPORT_D & ~(1<<IEC_DATA); else bit=IECPORT_D | (1<<IEC_DATA);
 while((FAST_SER_PORT_I & (1<<IEC_SRQ))==0); // wait for SRQ=1 (FAST_CLK)
 while((FAST_SER_PORT_I & (1<<IEC_SRQ))!=0); // wait for SRQ=0 (FAST_CLK)
 IECPORT_D=bit; //set DATA bit 4
 if ((byte & 0x08) != 0) bit=IECPORT_D & ~(1<<IEC_DATA); else bit=IECPORT_D | (1<<IEC_DATA);
 while((FAST_SER_PORT_I & (1<<IEC_SRQ))==0); // wait for SRQ=1 (FAST_CLK)
 while((FAST_SER_PORT_I & (1<<IEC_SRQ))!=0); // wait for SRQ=0 (FAST_CLK)
 IECPORT_D=bit; //set DATA bit 3
 if ((byte & 0x04) != 0) bit=IECPORT_D & ~(1<<IEC_DATA); else bit=IECPORT_D | (1<<IEC_DATA);
 while((FAST_SER_PORT_I & (1<<IEC_SRQ))==0); // wait for SRQ=1 (FAST_CLK)
 while((FAST_SER_PORT_I & (1<<IEC_SRQ))!=0); // wait for SRQ=0 (FAST_CLK)
 IECPORT_D=bit; //set DATA bit 2
 if ((byte & 0x02) != 0) bit=IECPORT_D & ~(1<<IEC_DATA); else bit=IECPORT_D | (1<<IEC_DATA);
 while((FAST_SER_PORT_I & (1<<IEC_SRQ))==0); // wait for SRQ=1 (FAST_CLK)
 while((FAST_SER_PORT_I & (1<<IEC_SRQ))!=0); // wait for SRQ=0 (FAST_CLK)
 IECPORT_D=bit; //set DATA bit 1
 if ((byte & 0x01) != 0) bit=IECPORT_D & ~(1<<IEC_DATA); else bit=IECPORT_D | (1<<IEC_DATA);
 while((FAST_SER_PORT_I & (1<<IEC_SRQ))==0); // wait for SRQ=1 (FAST_CLK)
 while((FAST_SER_PORT_I & (1<<IEC_SRQ))!=0); // wait for SRQ=0 (FAST_CLK)
 IECPORT_D=bit; //set DATA bit 0
 if ((byte & 0x01) != 0) bit=IECPORT_D & ~(1<<IEC_DATA); else bit=IECPORT_D | (1<<IEC_DATA);
 while((FAST_SER_PORT_I & (1<<IEC_SRQ))==0); // wait for SRQ=1 (FAST_CLK)
 delay_us(3); //skip 9th bit for MOS6522
 IECPORT_D &= ~(1<<IEC_DATA); // DATA=1
 }while(byte!=0);
}
fastbus_input(); // set SPI as slave input

}
//===================================================================================
void
send_buffer_to_fast_device()
{
	uint8_t i=0;

	while((i<20) && (iec_status)==0) //transfer to floppy input buffer
	{
		iec_send_byte(pgm_read_byte(&pgmLOADER7X[i++]));
	}
	iec_unlisten();
	if(iec_status==0)
	{
		iec_burst_write(1); //transfer 256 bytes to floppy $0300
	}
	else
	{
		iec_reset();
	}

}

//===================================================================================
void
send_cmd_to_fast_device()
{
uint8_t i;

IECPORT_D |= (1<<IEC_CLK); // CLK=0
receive_fast_byte(100); //handshake dummy
IECPORT_D &= ~(1<<IEC_CLK); // CLK=1

if(iec_status==0)
{
 i=2;
 fastbus_output();
 while((i<8) && (iec_status)==0) //send 6 bytes
 {
  delay_us(40);//40
  send_byte_fast(rxbuffer[i++]);

 }
delay_us(10);
}

}
//===================================================================================
//send 6 bytes in order to start drive utility prg 
void
send_cmd_to_device()
{
	uint8_t i;
	if (fastdevice==0)
	{
		IECPORT_D |= (1<<IEC_CLK); // CLK=0
		send_byte_fast_to_6522(0x55); //handshake dummy
		IECPORT_D &= ~(1<<IEC_CLK); // CLK=1

		if(iec_status==0)
		{
			i=2;
			while((i<8) && (iec_status==0)) //send 6 bytes
			{
				send_byte_fast_to_6522(rxbuffer[i++]);
			}
			send_byte_fast_to_6522(255); //timing dummy
		}
	}
	else
	{
		IECPORT_D |= (1<<IEC_CLK); // CLK=0
		receive_fast_byte(100); //handshake dummy
		IECPORT_D &= ~(1<<IEC_CLK); // CLK=1

		if(iec_status==0)
		{
			i=2;
			fastbus_output();
			while((i<8) && (iec_status)==0) //send 6 bytes
			{
				delay_us(40);//40
				send_byte_fast(rxbuffer[i++]);

			}
			delay_us(10);
		}
	}
}
//===================================================================================
//Timer1 timeout 35.67ms
void
get_CLK_lowtime()
{
	uint8_t timeouts=0;
		
TCNT1=0;
TIFR1=(1<<TOV1);
while((get_clk() != 0) && timeouts<60)// wait for CLK 1->0 transition (and spin-up + seek track time 2s)
 {
	 if(TIMEOUT_T1) {timeouts++;TIFR1=(1<<TOV1);}
 }
/*
while(get_clk() != 0);// wait for CLK 1->0 transition
*/
TCNT1=0;
TIFR1=(1<<TOV1);
timeouts=0;

 while((get_clk()==0) && timeouts<42) // wait for CLK 0->1 transition
 {
	 if(TIMEOUT_T1) {timeouts++;TIFR1=(1<<TOV1);}
 }


send_byte_to_PC(TCNT1L);
 send_byte_to_PC(TCNT1H);
 send_byte_to_PC(timeouts);
}
//===================================================================================
void
get_iec_level()
{
iec_status=0;
if(IECPORT_I & (1<<IEC_CLK)) iec_status |= 1;
if(IECPORT_I & (1<<IEC_DATA)) iec_status |= 2;
if(IECPORT_I & (1<<IEC_ATN)) iec_status |= 4;
if(IECPORT_I & (1<<IEC_RESET))  iec_status |= 8;
}
//===================================================================================
int
main (void)
{
uint16_t returnbytes;

init();

 for (;;)
 {
	  
  if(pvalid!=0) 
  {
   returnbytes=16;
   bytes2pc=0;
   iec_status=0;

   switch(pid)
    {
	case 1: //OPEN
	 iec_open();
     break;
	
	case 2: //CLOSE
	 iec_close();
     break;
	
	case 3: // send string to iec
	 iec_set_listener();
	 iec_send_string();
	 break;
	
	case 4: // get string from iec
	 returnbytes=256*rxbuffer[4] + rxbuffer[5];
	 iec_receive_string(returnbytes);
	 break;
	 
	case 5: // send one byte to iec
	 iec_send_byte(rxbuffer[2]);
	 break;
	 
	case 6: //set listener
     iec_set_listener();
	 break;
	 
	case 7: //set talker
	 iec_set_talker();
	 break;
	 
	case 8: //receive entire track; Floppy 1581->ATMEGA->PC
	 iec_set_listener();
	 iec_send_string();
	 iec_receive_track1581();
	 returnbytes=20*256+1;
	 break;
	
	case 9: //transmit 256 bytes; PC->ATMEGA->Floppy 1581
	 iec_set_listener();
	 iec_send_string();
	 iec_burst_write_track();
	 break;
	 
	case 10: //get IEC levels
	 get_iec_level();
	 break;
	 
	case 11: //dummy sectorbuffer
	 break;

	case 12: //speed upload 256 bytes to floppy
	 iec_set_listener();
	 if (fastdevice==0) send_buffer_to_slow_device(); else send_buffer_to_fast_device();
	 break;

	case 13: //send utility command to floppy (Memory Read, Write, Execute)
	 send_cmd_to_device();
	 send_byte_to_PC(receive_fast_byte(2500));
	 break;

	case 14: //read whole track or memory block
	 send_cmd_to_device();
	 returnbytes=rxbuffer[3]*256;
	 read_track(returnbytes);
	 returnbytes +=2;
	 break;
	 
	case 15: //write whole track
	 send_cmd_to_device();
	 if (fastdevice==0) write_track_slow_device(); else write_track_fast_device();
	 break;
	 
	case 16: //send utility command to floppy and measure timing via CLK low time
	send_cmd_to_device();
	get_CLK_lowtime();
	break;

	case 17: //read 1581 memory
	returnbytes=rxbuffer[5]*256;
	memory_read_1581(rxbuffer[4],rxbuffer[5]);
	break;
	
	case 170://ping
	 //send_string_RS232("SER2IEC V1.0");
	 break;
	case 255:
	 iec_reset();
	 returnbytes=0;
	 break;
	}
   frame_termination(returnbytes);
   pvalid=0;
   }	
 }
}


